## Fibonacci numbers
## f(n) = f(n-1) + f(n-2), f(0)= 0 and f(1) = 1
## 0,1,1,2,3,5,8,13,21,34,55,89, ...

from __future__ import print_function
from timeit import Timer

##non-recursive version
def calcFib(n):
    if n == 0:
        return 0
    elif n == 1:
        return 1
    else: #if n is greater than 1
        an_0 = 0 #term 0
        an_1 = 1 #term 1
        for i in range(1,n):
            temp = an_0
            an_0 = an_1
            an_1 = temp + an_1 #f(n) = f(n-1) + f(n-2)
        return an_1


##recursive version with recomutation
def calcFibRecur(n):
    if n == 0: #base case f(0)=0
        return 0
    elif n == 1: #base case f(1)=1
        return 1
    else: #if n is greater than 1, f(n) = f(n-1) + f(n-2)
        return calcFibRecur(n-1) + calcFibRecur(n-2)


##recursive version with no computation
##create a dictionary memo to keep all the returns in the recursion.
memo = {0:0, 1:1} ##the first two items in the dictionary are the base cases
def improvedFibRecur(n):
    if not n in memo:
        memo[n] = improvedFibRecur(n-1) + improvedFibRecur(n-2)
    return memo[n]


##compare the performance of function calcFib() and calcFibRecur()
##the unit of time is second
##each function will be executed by 3 times
##you can change the variable execute_times to set executed times for each function
execute_times = 3

print("{:<5} | {:<14} | {:<10} |  {:<10}".format("N", "calcFibRecur()", "calcFib()", "Percent"))
for i in range(30):
    #time the recursion version
    stmt = "calcFibRecur(" + str(i) +")"
    t1 = Timer(stmt,"from __main__ import calcFibRecur")
    time1 = t1.timeit(execute_times)

    #time the non-recursion version
    stmt = "calcFib(" + str(i) +")"
    t2 = Timer(stmt,"from __main__ import calcFib")
    time2 = t2.timeit(execute_times)
    
    print("n={:2}  |  {:7f}      |  {:7f}  |  {:7f}".format(i, time1, time2, time1/time2))

print()


##compare the performance of function calcFib() and improvedFibRecur()
print("{:<5} | {:<14} | {:<10} |  {:<10}".format("N", "improvedFibRecur()", "calcFib()", "Percent"))
for i in range(30):
    #time the recursion version
    stmt = "improvedFibRecur(" + str(i) +")"
    t1 = Timer(stmt,"from __main__ import improvedFibRecur")
    time1 = t1.timeit(execute_times)

    #time the non-recursion version
    stmt = "calcFib(" + str(i) +")"
    t2 = Timer(stmt,"from __main__ import calcFib")
    time2 = t2.timeit(execute_times)
    
    print("n={:2}  |  {:7f}          |  {:7f}  |  {:7f}".format(i, time1, time2, time1/time2))
    
